
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "config.h"
#include <flatpak-utils-private.h>
#include <flatpak.h>
#include <flatpak-enum-types.h>
#include <gio/gio.h>
/* enumerations from "flatpak-error.h" */
GType
flatpak_error_get_type (void)
{
  static gsize static_g_enum_type_id = 0;

  if (g_once_init_enter (&static_g_enum_type_id))
    {
      static const GEnumValue values[] = {
        { FLATPAK_ERROR_ALREADY_INSTALLED, "FLATPAK_ERROR_ALREADY_INSTALLED", "already-installed" },
        { FLATPAK_ERROR_NOT_INSTALLED, "FLATPAK_ERROR_NOT_INSTALLED", "not-installed" },
        { FLATPAK_ERROR_ONLY_PULLED, "FLATPAK_ERROR_ONLY_PULLED", "only-pulled" },
        { FLATPAK_ERROR_DIFFERENT_REMOTE, "FLATPAK_ERROR_DIFFERENT_REMOTE", "different-remote" },
        { FLATPAK_ERROR_ABORTED, "FLATPAK_ERROR_ABORTED", "aborted" },
        { FLATPAK_ERROR_SKIPPED, "FLATPAK_ERROR_SKIPPED", "skipped" },
        { FLATPAK_ERROR_NEED_NEW_FLATPAK, "FLATPAK_ERROR_NEED_NEW_FLATPAK", "need-new-flatpak" },
        { FLATPAK_ERROR_REMOTE_NOT_FOUND, "FLATPAK_ERROR_REMOTE_NOT_FOUND", "remote-not-found" },
        { FLATPAK_ERROR_RUNTIME_NOT_FOUND, "FLATPAK_ERROR_RUNTIME_NOT_FOUND", "runtime-not-found" },
        { FLATPAK_ERROR_DOWNGRADE, "FLATPAK_ERROR_DOWNGRADE", "downgrade" },
        { FLATPAK_ERROR_INVALID_REF, "FLATPAK_ERROR_INVALID_REF", "invalid-ref" },
        { FLATPAK_ERROR_INVALID_DATA, "FLATPAK_ERROR_INVALID_DATA", "invalid-data" },
        { FLATPAK_ERROR_UNTRUSTED, "FLATPAK_ERROR_UNTRUSTED", "untrusted" },
        { FLATPAK_ERROR_SETUP_FAILED, "FLATPAK_ERROR_SETUP_FAILED", "setup-failed" },
        { FLATPAK_ERROR_EXPORT_FAILED, "FLATPAK_ERROR_EXPORT_FAILED", "export-failed" },
        { FLATPAK_ERROR_REMOTE_USED, "FLATPAK_ERROR_REMOTE_USED", "remote-used" },
        { FLATPAK_ERROR_RUNTIME_USED, "FLATPAK_ERROR_RUNTIME_USED", "runtime-used" },
        { FLATPAK_ERROR_INVALID_NAME, "FLATPAK_ERROR_INVALID_NAME", "invalid-name" },
        { FLATPAK_ERROR_OUT_OF_SPACE, "FLATPAK_ERROR_OUT_OF_SPACE", "out-of-space" },
        { FLATPAK_ERROR_WRONG_USER, "FLATPAK_ERROR_WRONG_USER", "wrong-user" },
        { FLATPAK_ERROR_NOT_CACHED, "FLATPAK_ERROR_NOT_CACHED", "not-cached" },
        { FLATPAK_ERROR_REF_NOT_FOUND, "FLATPAK_ERROR_REF_NOT_FOUND", "ref-not-found" },
        { FLATPAK_ERROR_PERMISSION_DENIED, "FLATPAK_ERROR_PERMISSION_DENIED", "permission-denied" },
        { FLATPAK_ERROR_AUTHENTICATION_FAILED, "FLATPAK_ERROR_AUTHENTICATION_FAILED", "authentication-failed" },
        { FLATPAK_ERROR_NOT_AUTHORIZED, "FLATPAK_ERROR_NOT_AUTHORIZED", "not-authorized" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("FlatpakError"), values);
      g_once_init_leave (&static_g_enum_type_id, g_define_type_id);
    }

  return static_g_enum_type_id;
}
/* enumerations from "flatpak-installation.h" */
GType
flatpak_update_flags_get_type (void)
{
  static gsize static_g_flags_type_id = 0;

  if (g_once_init_enter (&static_g_flags_type_id))
    {
      static const GFlagsValue values[] = {
        { FLATPAK_UPDATE_FLAGS_NONE, "FLATPAK_UPDATE_FLAGS_NONE", "none" },
        { FLATPAK_UPDATE_FLAGS_NO_DEPLOY, "FLATPAK_UPDATE_FLAGS_NO_DEPLOY", "no-deploy" },
        { FLATPAK_UPDATE_FLAGS_NO_PULL, "FLATPAK_UPDATE_FLAGS_NO_PULL", "no-pull" },
        { FLATPAK_UPDATE_FLAGS_NO_STATIC_DELTAS, "FLATPAK_UPDATE_FLAGS_NO_STATIC_DELTAS", "no-static-deltas" },
        { FLATPAK_UPDATE_FLAGS_NO_PRUNE, "FLATPAK_UPDATE_FLAGS_NO_PRUNE", "no-prune" },
        { FLATPAK_UPDATE_FLAGS_NO_TRIGGERS, "FLATPAK_UPDATE_FLAGS_NO_TRIGGERS", "no-triggers" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_flags_register_static (g_intern_static_string ("FlatpakUpdateFlags"), values);
      g_once_init_leave (&static_g_flags_type_id, g_define_type_id);
    }

  return static_g_flags_type_id;
}
GType
flatpak_install_flags_get_type (void)
{
  static gsize static_g_flags_type_id = 0;

  if (g_once_init_enter (&static_g_flags_type_id))
    {
      static const GFlagsValue values[] = {
        { FLATPAK_INSTALL_FLAGS_NONE, "FLATPAK_INSTALL_FLAGS_NONE", "none" },
        { FLATPAK_INSTALL_FLAGS_NO_STATIC_DELTAS, "FLATPAK_INSTALL_FLAGS_NO_STATIC_DELTAS", "no-static-deltas" },
        { FLATPAK_INSTALL_FLAGS_NO_DEPLOY, "FLATPAK_INSTALL_FLAGS_NO_DEPLOY", "no-deploy" },
        { FLATPAK_INSTALL_FLAGS_NO_PULL, "FLATPAK_INSTALL_FLAGS_NO_PULL", "no-pull" },
        { FLATPAK_INSTALL_FLAGS_NO_TRIGGERS, "FLATPAK_INSTALL_FLAGS_NO_TRIGGERS", "no-triggers" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_flags_register_static (g_intern_static_string ("FlatpakInstallFlags"), values);
      g_once_init_leave (&static_g_flags_type_id, g_define_type_id);
    }

  return static_g_flags_type_id;
}
GType
flatpak_uninstall_flags_get_type (void)
{
  static gsize static_g_flags_type_id = 0;

  if (g_once_init_enter (&static_g_flags_type_id))
    {
      static const GFlagsValue values[] = {
        { FLATPAK_UNINSTALL_FLAGS_NONE, "FLATPAK_UNINSTALL_FLAGS_NONE", "none" },
        { FLATPAK_UNINSTALL_FLAGS_NO_PRUNE, "FLATPAK_UNINSTALL_FLAGS_NO_PRUNE", "no-prune" },
        { FLATPAK_UNINSTALL_FLAGS_NO_TRIGGERS, "FLATPAK_UNINSTALL_FLAGS_NO_TRIGGERS", "no-triggers" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_flags_register_static (g_intern_static_string ("FlatpakUninstallFlags"), values);
      g_once_init_leave (&static_g_flags_type_id, g_define_type_id);
    }

  return static_g_flags_type_id;
}
GType
flatpak_launch_flags_get_type (void)
{
  static gsize static_g_flags_type_id = 0;

  if (g_once_init_enter (&static_g_flags_type_id))
    {
      static const GFlagsValue values[] = {
        { FLATPAK_LAUNCH_FLAGS_NONE, "FLATPAK_LAUNCH_FLAGS_NONE", "none" },
        { FLATPAK_LAUNCH_FLAGS_DO_NOT_REAP, "FLATPAK_LAUNCH_FLAGS_DO_NOT_REAP", "do-not-reap" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_flags_register_static (g_intern_static_string ("FlatpakLaunchFlags"), values);
      g_once_init_leave (&static_g_flags_type_id, g_define_type_id);
    }

  return static_g_flags_type_id;
}
GType
flatpak_query_flags_get_type (void)
{
  static gsize static_g_flags_type_id = 0;

  if (g_once_init_enter (&static_g_flags_type_id))
    {
      static const GFlagsValue values[] = {
        { FLATPAK_QUERY_FLAGS_NONE, "FLATPAK_QUERY_FLAGS_NONE", "none" },
        { FLATPAK_QUERY_FLAGS_ONLY_CACHED, "FLATPAK_QUERY_FLAGS_ONLY_CACHED", "only-cached" },
        { FLATPAK_QUERY_FLAGS_ONLY_SIDELOADED, "FLATPAK_QUERY_FLAGS_ONLY_SIDELOADED", "only-sideloaded" },
        { FLATPAK_QUERY_FLAGS_ALL_ARCHES, "FLATPAK_QUERY_FLAGS_ALL_ARCHES", "all-arches" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_flags_register_static (g_intern_static_string ("FlatpakQueryFlags"), values);
      g_once_init_leave (&static_g_flags_type_id, g_define_type_id);
    }

  return static_g_flags_type_id;
}
GType
flatpak_storage_type_get_type (void)
{
  static gsize static_g_enum_type_id = 0;

  if (g_once_init_enter (&static_g_enum_type_id))
    {
      static const GEnumValue values[] = {
        { FLATPAK_STORAGE_TYPE_DEFAULT, "FLATPAK_STORAGE_TYPE_DEFAULT", "default" },
        { FLATPAK_STORAGE_TYPE_HARD_DISK, "FLATPAK_STORAGE_TYPE_HARD_DISK", "hard-disk" },
        { FLATPAK_STORAGE_TYPE_SDCARD, "FLATPAK_STORAGE_TYPE_SDCARD", "sdcard" },
        { FLATPAK_STORAGE_TYPE_MMC, "FLATPAK_STORAGE_TYPE_MMC", "mmc" },
        { FLATPAK_STORAGE_TYPE_NETWORK, "FLATPAK_STORAGE_TYPE_NETWORK", "network" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("FlatpakStorageType"), values);
      g_once_init_leave (&static_g_enum_type_id, g_define_type_id);
    }

  return static_g_enum_type_id;
}
/* enumerations from "flatpak-portal-error.h" */
GType
flatpak_portal_error_get_type (void)
{
  static gsize static_g_enum_type_id = 0;

  if (g_once_init_enter (&static_g_enum_type_id))
    {
      static const GEnumValue values[] = {
        { FLATPAK_PORTAL_ERROR_FAILED, "FLATPAK_PORTAL_ERROR_FAILED", "failed" },
        { FLATPAK_PORTAL_ERROR_INVALID_ARGUMENT, "FLATPAK_PORTAL_ERROR_INVALID_ARGUMENT", "invalid-argument" },
        { FLATPAK_PORTAL_ERROR_NOT_FOUND, "FLATPAK_PORTAL_ERROR_NOT_FOUND", "not-found" },
        { FLATPAK_PORTAL_ERROR_EXISTS, "FLATPAK_PORTAL_ERROR_EXISTS", "exists" },
        { FLATPAK_PORTAL_ERROR_NOT_ALLOWED, "FLATPAK_PORTAL_ERROR_NOT_ALLOWED", "not-allowed" },
        { FLATPAK_PORTAL_ERROR_CANCELLED, "FLATPAK_PORTAL_ERROR_CANCELLED", "cancelled" },
        { FLATPAK_PORTAL_ERROR_WINDOW_DESTROYED, "FLATPAK_PORTAL_ERROR_WINDOW_DESTROYED", "window-destroyed" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("FlatpakPortalError"), values);
      g_once_init_leave (&static_g_enum_type_id, g_define_type_id);
    }

  return static_g_enum_type_id;
}
/* enumerations from "flatpak-ref.h" */
GType
flatpak_ref_kind_get_type (void)
{
  static gsize static_g_enum_type_id = 0;

  if (g_once_init_enter (&static_g_enum_type_id))
    {
      static const GEnumValue values[] = {
        { FLATPAK_REF_KIND_APP, "FLATPAK_REF_KIND_APP", "app" },
        { FLATPAK_REF_KIND_RUNTIME, "FLATPAK_REF_KIND_RUNTIME", "runtime" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("FlatpakRefKind"), values);
      g_once_init_leave (&static_g_enum_type_id, g_define_type_id);
    }

  return static_g_enum_type_id;
}
/* enumerations from "flatpak-remote.h" */
GType
flatpak_remote_type_get_type (void)
{
  static gsize static_g_enum_type_id = 0;

  if (g_once_init_enter (&static_g_enum_type_id))
    {
      static const GEnumValue values[] = {
        { FLATPAK_REMOTE_TYPE_STATIC, "FLATPAK_REMOTE_TYPE_STATIC", "static" },
        { FLATPAK_REMOTE_TYPE_USB, "FLATPAK_REMOTE_TYPE_USB", "usb" },
        { FLATPAK_REMOTE_TYPE_LAN, "FLATPAK_REMOTE_TYPE_LAN", "lan" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("FlatpakRemoteType"), values);
      g_once_init_leave (&static_g_enum_type_id, g_define_type_id);
    }

  return static_g_enum_type_id;
}
/* enumerations from "flatpak-transaction.h" */
GType
flatpak_transaction_operation_type_get_type (void)
{
  static gsize static_g_enum_type_id = 0;

  if (g_once_init_enter (&static_g_enum_type_id))
    {
      static const GEnumValue values[] = {
        { FLATPAK_TRANSACTION_OPERATION_INSTALL, "FLATPAK_TRANSACTION_OPERATION_INSTALL", "install" },
        { FLATPAK_TRANSACTION_OPERATION_UPDATE, "FLATPAK_TRANSACTION_OPERATION_UPDATE", "update" },
        { FLATPAK_TRANSACTION_OPERATION_INSTALL_BUNDLE, "FLATPAK_TRANSACTION_OPERATION_INSTALL_BUNDLE", "install-bundle" },
        { FLATPAK_TRANSACTION_OPERATION_UNINSTALL, "FLATPAK_TRANSACTION_OPERATION_UNINSTALL", "uninstall" },
        { FLATPAK_TRANSACTION_OPERATION_LAST_TYPE, "FLATPAK_TRANSACTION_OPERATION_LAST_TYPE", "last-type" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("FlatpakTransactionOperationType"), values);
      g_once_init_leave (&static_g_enum_type_id, g_define_type_id);
    }

  return static_g_enum_type_id;
}
GType
flatpak_transaction_error_details_get_type (void)
{
  static gsize static_g_flags_type_id = 0;

  if (g_once_init_enter (&static_g_flags_type_id))
    {
      static const GFlagsValue values[] = {
        { FLATPAK_TRANSACTION_ERROR_DETAILS_NON_FATAL, "FLATPAK_TRANSACTION_ERROR_DETAILS_NON_FATAL", "fatal" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_flags_register_static (g_intern_static_string ("FlatpakTransactionErrorDetails"), values);
      g_once_init_leave (&static_g_flags_type_id, g_define_type_id);
    }

  return static_g_flags_type_id;
}
GType
flatpak_transaction_result_get_type (void)
{
  static gsize static_g_flags_type_id = 0;

  if (g_once_init_enter (&static_g_flags_type_id))
    {
      static const GFlagsValue values[] = {
        { FLATPAK_TRANSACTION_RESULT_NO_CHANGE, "FLATPAK_TRANSACTION_RESULT_NO_CHANGE", "change" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_flags_register_static (g_intern_static_string ("FlatpakTransactionResult"), values);
      g_once_init_leave (&static_g_flags_type_id, g_define_type_id);
    }

  return static_g_flags_type_id;
}
GType
flatpak_transaction_remote_reason_get_type (void)
{
  static gsize static_g_enum_type_id = 0;

  if (g_once_init_enter (&static_g_enum_type_id))
    {
      static const GEnumValue values[] = {
        { FLATPAK_TRANSACTION_REMOTE_GENERIC_REPO, "FLATPAK_TRANSACTION_REMOTE_GENERIC_REPO", "generic-repo" },
        { FLATPAK_TRANSACTION_REMOTE_RUNTIME_DEPS, "FLATPAK_TRANSACTION_REMOTE_RUNTIME_DEPS", "runtime-deps" },
        { 0, NULL, NULL }
      };
      GType g_define_type_id =
        g_enum_register_static (g_intern_static_string ("FlatpakTransactionRemoteReason"), values);
      g_once_init_leave (&static_g_enum_type_id, g_define_type_id);
    }

  return static_g_enum_type_id;
}

/* Generated data ends here */

