
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifndef __FLATPAK_ENUM_TYPES_H__
#define __FLATPAK_ENUM_TYPES_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "flatpak-error.h" */
FLATPAK_EXTERN GType flatpak_error_get_type (void);
#define FLATPAK_TYPE_ERROR (flatpak_error_get_type ())

/* enumerations from "flatpak-installation.h" */
FLATPAK_EXTERN GType flatpak_update_flags_get_type (void);
#define FLATPAK_TYPE_UPDATE_FLAGS (flatpak_update_flags_get_type ())
FLATPAK_EXTERN GType flatpak_install_flags_get_type (void);
#define FLATPAK_TYPE_INSTALL_FLAGS (flatpak_install_flags_get_type ())
FLATPAK_EXTERN GType flatpak_uninstall_flags_get_type (void);
#define FLATPAK_TYPE_UNINSTALL_FLAGS (flatpak_uninstall_flags_get_type ())
FLATPAK_EXTERN GType flatpak_launch_flags_get_type (void);
#define FLATPAK_TYPE_LAUNCH_FLAGS (flatpak_launch_flags_get_type ())
FLATPAK_EXTERN GType flatpak_query_flags_get_type (void);
#define FLATPAK_TYPE_QUERY_FLAGS (flatpak_query_flags_get_type ())
FLATPAK_EXTERN GType flatpak_storage_type_get_type (void);
#define FLATPAK_TYPE_STORAGE_TYPE (flatpak_storage_type_get_type ())

/* enumerations from "flatpak-portal-error.h" */
FLATPAK_EXTERN GType flatpak_portal_error_get_type (void);
#define FLATPAK_TYPE_PORTAL_ERROR (flatpak_portal_error_get_type ())

/* enumerations from "flatpak-ref.h" */
FLATPAK_EXTERN GType flatpak_ref_kind_get_type (void);
#define FLATPAK_TYPE_REF_KIND (flatpak_ref_kind_get_type ())

/* enumerations from "flatpak-remote.h" */
FLATPAK_EXTERN GType flatpak_remote_type_get_type (void);
#define FLATPAK_TYPE_REMOTE_TYPE (flatpak_remote_type_get_type ())

/* enumerations from "flatpak-transaction.h" */
FLATPAK_EXTERN GType flatpak_transaction_operation_type_get_type (void);
#define FLATPAK_TYPE_TRANSACTION_OPERATION_TYPE (flatpak_transaction_operation_type_get_type ())
FLATPAK_EXTERN GType flatpak_transaction_error_details_get_type (void);
#define FLATPAK_TYPE_TRANSACTION_ERROR_DETAILS (flatpak_transaction_error_details_get_type ())
FLATPAK_EXTERN GType flatpak_transaction_result_get_type (void);
#define FLATPAK_TYPE_TRANSACTION_RESULT (flatpak_transaction_result_get_type ())
FLATPAK_EXTERN GType flatpak_transaction_remote_reason_get_type (void);
#define FLATPAK_TYPE_TRANSACTION_REMOTE_REASON (flatpak_transaction_remote_reason_get_type ())
G_END_DECLS

#endif /* __GIO_ENUM_TYPES_H__ */

/* Generated data ends here */

